/*
 *  galign.cpp
 *  PHunterLib
 *
 *  Created by Tiayyba Riaz on 6/7/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

//=============================================================================
// Module:        galign.cpp
// Project:       Cubic Project - Calculation of melting temperature and free
//                energy of two DNA strands 
// Type:          implementation - Thermodynamic Alignment.
// Language:      c++
// Compiler:      microsoft visual c++ 6.0, unix/linux gcc
// System/OS:     Windows 32, Sun solaris, Linux, other unix systems (untested)
// Database:      none
// Description:   class GAlign - Thermodynamic Alignment Algorithm
// Author:        leber
// Date:          01/2002 - 02/2002
// Copyright:     (c) L. Kaderali & M. Leber, 01/2002 - 02/2002
//
// Revision History
// $ 00sep04 LK : created 
// $ 00dec30 LK : changed to do local alignment of probe against
//                one entire sequence 
// $ 01feb07 LK : optimized!
// $ 01aug06 LK : corrected, included salt and concentration input;
// $              made true local alignment (problem with initial mismatches!) 
// #$
//=============================================================================

#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "galign.h"


#define nndH(a,b,c) (*(galn->dH+((b)*(galn->maxseq1len+1)*3+(a)*3+(c))))
#define nndS(a,b,c) (*(galn->dS+((b)*(galn->maxseq1len+1)*3+(a)*3+(c))))



///////////////////////////////////////////////////////////////////////////////
// Construction/Destruction
///////////////////////////////////////////////////////////////////////////////

//------------------------ Constructor ----------------------------------------
// Initialize Thermalign Object. This will allocate memory for the dynamic
// programming matrices for both entropy and enthalpy of the present
// alignment. The max length of the two sequences to align has to be passed 
// to the routine. Also, an initialized CNNParams Object must be passed to 
// CThermAlign. Object will be initialized only once, sequences can be
// passed to it later.

//int s1lmax,s2lmax;  // global requiered for DEBUGGING

void galn_initGAlign(GGAlign galn, int ALength, int BLength, PNNParams myNNParams)
{
    // Create dynamic programming matrices for entropy and enthalpy.
    // The first two dimensions are for the two strings respectively, the
    // third dimensions differentiates the alignment according to its end:
    // type 0 Alignments will be those that align x(i+1) with y(j+1),
    // type 1 Alignments align x(i+1) with a gap
    // type 2 Alignments align y(i+1) with a gap
    // All three of these types must separately be considered to select the
    // optimum alignment in the next iteration.
    // The alignment algorithm proceeds analogous to the well known
    // Smith-Waterman algorithm for global alignments.

    galn->dH = (float *) calloc((ALength+1)*(BLength+1)*3, sizeof (float));
    galn->dS = (float *) calloc((ALength+1)*(BLength+1)*3, sizeof (float));
//    s1lmax=ALength;
//    s2lmax=BLength;

	if ((galn->dH==NULL)||(galn->dS==NULL))
	{
		printf("Fatal Error: Out of memory!\n");
		exit(-1);
	}

    galn->GNNParams = myNNParams;
	galn->seq1len = ALength;
	galn->maxseq1len = galn->seq1len;
	galn->seq2len = BLength;

	memset(galn->dH,0,(ALength+1)*(BLength+1)*3*sizeof(float));
	memset(galn->dS,0,(ALength+1)*(BLength+1)*3*sizeof(float));

    // set dH / dS entries all to zero!
    #ifdef _pack
	    // this is FAST, but works only if #pragma pack(1) has been set...
    #else
		// the following is slower, but works also when padding bytes have 
		// been inserted in the array for alignment by the compiler...
		// time is not really critical here, as this initialization is done
		// only once...
/*		int a, b;
		for ( a=0;a<=ALength;a++)
			for ( b=0;b<=BLength;b++)
			{
				nndH(a,b,0)=0.0f;
				nndH(a,b,1)=0.0f;
				nndH(a,b,2)=0.0f;
				nndS(a,b,0)=0.0f;
				nndS(a,b,1)=0.0f;
				nndS(a,b,2)=0.0f;
			}*/
    #endif
    galn_InitBorder(galn);  // also need to do only once...
//    forbidden_entropy = (-(GNNParams->R)*(float)log((GNNParams->Ct)/4.0f));
    return;
}

//------------------------ Destructor -----------------------------------------
// Free memory

void galn_finiGAlign(GGAlign galn)
{
	free (galn->dH);
	free (galn->dS);
    return;
}



///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   void galn_InitStrings(char* s1, 
//                                            char* s2,
//                                            int s1len,
//                                            int s2len);
//
//  PURPOSE:    Initialize strings for alignment table!
//
//  PARAMETERS:
//      s1:    Target sequence
//      s2:    Probe sequence
//      s1len: Target sequence length
//      s2len: Probe sequence length
//
//  RETURN VALUE:
//      void
//
//  REVISION HISTORY
//  $ 01jan03 LK : created
//  $ 01feb07 LK : removed maximum local stuff
//  #$

void galn_InitStrings(GGAlign galn, char* s1,char* s2,int s1len ,int s2len,int minxpos)
{
	galn->seq1=s1; galn->seq1len=s1len;
	galn->seq2=s2; galn->seq2len=s2len;
	nparam_UpdateParams(galn->GNNParams, s1,s2);
/* lk01feb07: removed maxloc-stuff
	// if maximum local is not in reused subtable, need to reset!
	if ((maxlocj>=minxpos)||(maxloci!=seq1len))
	{
		maxloci=0; maxlocj=0; maxlocg=0;
	}
*/
//    InitBorder();   //lk01feb07: removed, as it suffices if this is done once in constructor
	return;
}


///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   void galn_InitBorder();
//
//  PURPOSE:    Initialize the lower and left border of the dynamic
//              programming table. 
//
//  PARAMETERS:
//      none
//
//  RETURN VALUE:
//      void
//
//  REVISION HISTORY
//  $ 00sep06 LK : created
//  $ 00dec30 LK : complete revision - simply init table now, no
//                 more saving etc...
//  $ 01feb07 LK : removed local alignment stuff and optimized
//  #$

void galn_InitBorder(GGAlign galn)
{
	// Initialization for new alignment!
	galn->maxlocg = -99999;   // maximum melting temperature found (kelvin)
	galn->maxloci  = 0;
	galn->maxlocj  = 0;
	galn->maxloct  = 0;
	int i, j;
	
	for ( i=0; i<=galn->seq1len; i++)
        {
		nndH(i,0,0) = 0.0f;	nndH(i,0,1) = 0.0f;	nndH(i,0,2) = 0.0f;
		nndS(i,0,0) = nparam_GetInitialEntropy(galn->GNNParams);	
                nndS(i,0,1) = nparam_GetInitialEntropy(galn->GNNParams);	
                nndS(i,0,2) = nparam_GetInitialEntropy(galn->GNNParams);     
        }     
        for  ( j=1; j<=galn->seq2len; j++)     {
		nndH(0,j,0) = 0.0f;	nndH(0,j,1) = 0.0f;	nndH(0,j,2) = 0.0f;
		nndS(0,j,0) = nparam_GetInitialEntropy(galn->GNNParams); 
                nndS(0,j,1) = nparam_GetInitialEntropy(galn->GNNParams);
                nndS(0,j,2) = nparam_GetInitialEntropy(galn->GNNParams);  
        }

    return;
}


///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   galn_CalculateTable();
//
//  PURPOSE:    Compute the entire dynamic programming table cell by cell.
//              Requieres that InitBorder has been called.
//              Computations will be done in the usual rowwise manner.
//
//  PARAMETERS:
//      int startRow:  Row that computation shall commence at!
//
//  RETURN VALUE:
//      void - Data will be written directly to dH and dS in this object 
//
//  REVISION HISTORY
//  $              00sep06 : created LK
//                 00dec31 : modified to begin at a specified row!
//  #$

void galn_CalculateTable(GGAlign galn, int startRow)
{
//	#define forbidden_enthalpy 1000000000000000000.0f
	register int i;
	register int iminusone;
	register int j;
	register int jminusone;
	// Determine previous character!
	register char prevchari;
	register char prevcharj;
	register char seq1char;
	register char seq2char;
	register float entropy1;
	register float entropy2;
	register float entropy3;
	register float enthalpy1;
	register float enthalpy2;
	register float enthalpy3;
	register float g1;
	register float g2;
	register float g3;
	float maxg;
//    ofstream dboutstream("debug.out");
//    assert(startRow>0);
    for (j=startRow; j<=galn->seq2len; j++)
	{
		if (j>1)
			prevcharj = galn->seq2[j-2];
		else
			prevcharj = 0;
		jminusone = j - 1;
		seq2char = galn->seq2[jminusone];
        for (i=1; i<=galn->seq1len; i++)
        {
		   // local variables
			if (i>1)
				prevchari = galn->seq1[i-2];
			else
				prevchari = 0;
			iminusone = i - 1;
			seq1char=galn->seq1[iminusone];

			// --------------- Upper cell: Alignment of seq1[i] with seg2[j]
			entropy1 = nndS(iminusone,jminusone,0);
			enthalpy1 = nndH(iminusone,jminusone,0) ;
			entropy2 = nndS(iminusone,jminusone,1);
			enthalpy2 = nndH(iminusone,jminusone,1);
			entropy3 = nndS(iminusone,jminusone,2);
			enthalpy3 = nndH(iminusone,jminusone,2);
			if (enthalpy1<forbidden_enthalpy)
			{
			     entropy1 += nparam_GetEntropy(galn->GNNParams, prevchari,seq1char, prevcharj, seq2char);
			     enthalpy1 +=nparam_GetEnthalpy(galn->GNNParams, prevchari, seq1char, prevcharj, seq2char);

		        } 			
                        if (enthalpy2<forbidden_enthalpy)
			{
			     enthalpy2 +=nparam_GetEnthalpy(galn->GNNParams, prevchari,seq1char, 0, seq2char); // 0 is gap!
                             entropy2 += nparam_GetEntropy(galn->GNNParams, prevchari,seq1char, 0, seq2char);
 			} 			
                        if(enthalpy3<forbidden_enthalpy) 			
                        {
		             enthalpy3 += nparam_GetEnthalpy(galn->GNNParams, 0, seq1char,prevcharj, seq2char);
                             entropy3 +=nparam_GetEntropy(galn->GNNParams, 0, seq1char, prevcharj, seq2char);
 	
                  	} 			
                        // choose optimum combination
			g1 = nparam_CalcG(galn->GNNParams, entropy1, enthalpy1);
			g2 = nparam_CalcG(galn->GNNParams, entropy2, enthalpy2);
			g3 = nparam_CalcG(galn->GNNParams, entropy3, enthalpy3);
			if ((g1>=g2)&&(g1>=g3))
			{
				nndS(i,j,0) = entropy1;
				nndH(i,j,0) = enthalpy1;
				maxg=g1;
			}
			else if (g2>=g3)
			{
				nndS(i,j,0) = entropy2;
				nndH(i,j,0) = enthalpy2;
				maxg=g2;
			}
			else
			{
				nndS(i,j,0) = entropy3;
				nndH(i,j,0) = enthalpy3;
				maxg=g3;
			}
		    // set to zero if alignment so far is mismatches only (tm will be zero -> no need to reset maxg)!
                        if(nndH(i,j,0)==0)
                            nndS(i,j,0)=nparam_GetInitialEntropy(galn->GNNParams);
			// check if local maximum found!
			if (((i==galn->seq1len)||(j==galn->seq2len))&&(maxg>=galn->maxlocg))
			{
				galn->maxlocg=maxg;
				galn->maxloci=i;
				galn->maxlocj=j;
				galn->maxloct=0;
			}
		    // set to zero if alignment so far is mismatches only!
			if (nndH(i,j,0)==0)
			    nndS(i,j,0)=nparam_GetInitialEntropy(galn->GNNParams);

			// --------------- Middle cell: Alignment of seq1[i] with '-'
			// following changes lk 01jan08
			// we do not allow $- in the beginning, therefore set to forbidden if
			// j=1
			if (j==1)
			{
				enthalpy1=forbidden_enthalpy; enthalpy2=forbidden_enthalpy;
				entropy1=forbidden_entropy; entropy2=forbidden_entropy;
			}
			// also, disallow -$ in the end, therefore forbid if j=seq2len-1
			else if (j==galn->seq2len-1)
			{
				enthalpy1=forbidden_enthalpy; enthalpy2=forbidden_enthalpy;
				entropy1=forbidden_entropy; entropy2=forbidden_entropy;
			}
			else
			{
				// end lk01jan08
				// -- entropy
				entropy1 = nndS(iminusone,j,0);
				entropy2 = nndS(iminusone,j,1);
				enthalpy1 = nndH(iminusone,j,0);
				enthalpy2 = nndH(iminusone,j,1);
				if (enthalpy1<forbidden_enthalpy)
				{
				    entropy1+=nparam_GetEntropy(galn->GNNParams, prevchari, seq1char, seq2char, 0);
 	                            enthalpy1 +=nparam_GetEnthalpy(galn->GNNParams, prevchari,seq1char, seq2char, 0);
                                }  	 
                                if (enthalpy2<forbidden_enthalpy)
                                {
	 	 		    entropy2 +=nparam_GetEntropy(galn->GNNParams, prevchari, seq1char, 0, 0); 
                                    enthalpy2 +=nparam_GetEnthalpy(galn->GNNParams, prevchari, seq1char, 0, 0); 		
	 			}
                        }	 	
                        // -- choose optimum combination 
                        g1 = nparam_CalcG(galn->GNNParams, entropy1, enthalpy1);
                        g2 = nparam_CalcG(galn->GNNParams, entropy2, enthalpy2);  
                        if (g1>=g2)
                        { 	
                                    nndS(i,j,1)=entropy1; 	
                                    nndH(i,j,1)=enthalpy1; 	
                                    maxg=g1;
                        }
			else
			{
				nndS(i,j,1)=entropy2;
				nndH(i,j,1)=enthalpy2;
				maxg=g2;
			}
			// check if local maximum found!
			if (((i==galn->seq1len)||(j==galn->seq2len))&&(maxg>=galn->maxlocg))
			{
				galn->maxlocg=maxg;
				galn->maxloci=i;
				galn->maxlocj=j;
				galn->maxloct=1;
			}
		    // set to zero if alignment so far is mismatches only!
			if (nndH(i,j,1)==0)
			    nndS(i,j,1)=nparam_GetInitialEntropy(galn->GNNParams);
		
			// --------------- Lower cell: Alignment of '-' with seq2[j]
			// following changes lk 01jan08
			// we do not allow $- in the beginning, therefore set to forbidden if
			// i=1
			if (i==1)
			{
				enthalpy1=forbidden_enthalpy; enthalpy2=forbidden_enthalpy;
				entropy1=forbidden_entropy; entropy2=forbidden_entropy;
			}
			// also, disallow -$ in the end, therefore forbid if i=seq1len-1
			else if (i==galn->seq1len-1)
			{
				enthalpy1=forbidden_enthalpy; enthalpy2=forbidden_enthalpy;
				entropy1=forbidden_entropy; entropy2=forbidden_entropy;
			}
			else
			{
			// end lk01jan08
				entropy1 = nndS(i,jminusone,0);
				entropy2 = nndS(i,jminusone,2);
				enthalpy1 = nndH(i,jminusone,0);
				enthalpy2 = nndH(i,jminusone,2);
				if (enthalpy1<forbidden_enthalpy)
				{
					entropy1+=nparam_GetEntropy(galn->GNNParams, seq1char, 0, prevcharj, seq2char);
 				        enthalpy1 +=nparam_GetEnthalpy(galn->GNNParams, seq1char, 0, prevcharj, seq2char); 
                                } 		 	
                                if(enthalpy2<forbidden_enthalpy)
 				{ 	
                                        entropy2 += nparam_GetEntropy(galn->GNNParams, 0,0, prevcharj, seq2char); 
                                        enthalpy2 +=nparam_GetEnthalpy(galn->GNNParams, 0, 0, prevcharj, seq2char);
                                } 	 	
                        } 			
                        // -- choose optimum combination 	
                        g1 = nparam_CalcG(galn->GNNParams, entropy1, enthalpy1); 	
                        g2 = nparam_CalcG(galn->GNNParams, entropy2, enthalpy2); 	
	                if (g1>=g2)
                        {
				nndS(i,j,2)=entropy1;
				nndH(i,j,2)=enthalpy1;
				maxg=g1;
			}
			else
			{
				nndS(i,j,2)=entropy2;
				nndH(i,j,2)=enthalpy2;
				maxg=g2;
			}
			// check if local maximum found!
			if (((i==galn->seq1len)||(j==galn->seq2len))&&(maxg>=galn->maxlocg))
			{
				galn->maxlocg=maxg;
				galn->maxloci=i;
				galn->maxlocj=j;
				galn->maxloct=2;
			}
		    // set to zero if alignment so far is mismatches only!
			if (nndH(i,j,2)==0)
			    nndS(i,j,2)=nparam_GetInitialEntropy(galn->GNNParams);
			
		
		/*
            dboutstream<<"After cell ("<<i<<","<<j<<"):"<<endl;
            PrintDPTable(dboutstream);
            dboutstream<<endl<<flush;
		*/
        }
	}
    //printEnthalpyTable(0);
    //printEntropyTable(0);
    return;
}

void galn_printEnthalpyTable(GGAlign galn, int level)
{
	int i, j;
	for( i=0;i<=galn->seq1len;i++) 
	{
		for( j=0;j<=galn->seq2len;j++)
		{
			printf ("%f ", nndH(i,j,level));
		}
		printf ("\n");
	}
	printf ("\n");
}

void galn_printEntropyTable(GGAlign galn, int level)
{
	int i, j;
	for( i=0;i<=galn->seq1len;i++) 
	{
		for( j=0;j<=galn->seq2len;j++)
		{
			printf ("%f ", nndH(i,j,level));
		}
		printf ("\n");
	}
	printf ("\n");
}


///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   float GetEntropy(int i, int j);
//
//  PURPOSE:    Return Entropy for given cell
//
//  PARAMETERS:
//      i, j    - Cell coordinates
//
//  RETURN VALUE:
//      float  - Entropy (dS) 
//
//  REVISION HISTORY
//  $              00sep23 : created LK
//  #$

/* inline */ float galn_GetEntropy(GGAlign galn, int i, int j)
{
	float entropy;

	// first, need to determine optimum values for dG and dH
	float tm0, tm1, tm2;
	tm0 = nparam_CalcTM(nndS(i,j,0),nndH(i,j,0));
	tm1 = nparam_CalcTM(nndS(i,j,1),nndH(i,j,1));
	tm2 = nparam_CalcTM(nndS(i,j,2),nndH(i,j,2));
	if ((tm0>=tm1)&&(tm0>=tm2))
		entropy = nndS(i,j,0);
	else if ((tm1>=tm0)&&(tm1>=tm2))
		entropy = nndS(i,j,1);
	else
		entropy = nndS(i,j,2);

	return entropy;
}


///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   float GetEnthalpy(int i, int j);
//
//  PURPOSE:    Return Enthalpy for given cell
//
//  PARAMETERS:
//      i, j    - Cell coordinates
//
//  RETURN VALUE:
//      float  - Enthalpy (dS) 
//
//  REVISION HISTORY
//  $              00sep23 : created LK
//  #$

/* inline */ float galn_GetEnthalpy(GGAlign galn, int i, int j)
{
	float enthalpy;

	// first, need to determine optimum values for dG and dH
	float tm0, tm1, tm2;
	tm0 = nparam_CalcTM(nndS(i,j,0),nndH(i,j,0));
	tm1 = nparam_CalcTM(nndS(i,j,1),nndH(i,j,1));
	tm2 = nparam_CalcTM(nndS(i,j,2),nndH(i,j,2));
	if ((tm0>=tm1)&&(tm0>=tm2))
		enthalpy = nndH(i,j,0);
	else if ((tm1>=tm0)&&(tm1>=tm2))
		enthalpy = nndH(i,j,1);
	else
		enthalpy = nndH(i,j,2);

	return enthalpy;
}



///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   float GetFreeEnergyK(int i, int j, float t);
//
//  PURPOSE:    Return free Energy for given cell at given temperature
//
//  PARAMETERS:
//      i, j    - Cell coordinates
//      t       - Temperature in Kelvin
//
//  RETURN VALUE:
//      float  - free Energy value (dG) at given temperature (Kelvin)
//
//  REVISION HISTORY
//  $              00sep06 : created LK
//  #$

float galn_GetFreeEnergyK(GGAlign galn, int i, int j, float t)
{
	// dG = dH - T * dS 

	float entropy;
	float enthalpy;

	// first, need to determine optimum values for dG and dH
	entropy = -galn_GetEntropy(galn, i,j);
	enthalpy = -galn_GetEnthalpy(galn, i,j);

	return enthalpy - t * entropy;
}


///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   float GetFreeEnergyC(int i, int j, int t);
//
//  PURPOSE:    Return free Energy for given cell at given temperature
//
//  PARAMETERS:
//      i, j    - Cell coordinates
//      t       - Temperature in Celsius
//
//  RETURN VALUE:
//      float  - free Energy value (dG) at melting temperature
//
//  REVISION HISTORY
//  $              00sep06 : created LK
//  #$

float galn_GetFreeEnergyC(GGAlign galn, int i, int j, float t)
{
	// dG = dH - T * dS

	float entropy;
	float enthalpy;

	// first, need to determine optimum values for dG and dH
	entropy = galn_GetEntropy(galn, i,j);
	enthalpy = galn_GetEnthalpy(galn, i,j);

	return enthalpy - (t + 273.15f) * entropy;
}

///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   float galn_GetMeltingTempC(int i, int j)
//
//  PURPOSE:    Calculate TM for given cell. Return in Celsius.
//
//  PARAMETERS:
//      i, j    - Cell coordinates
//
//  RETURN VALUE:
//      float  - Melting Temperature TM
//
//  REVISION HISTORY
//  $              00sep06 : created LK
//  #$

float galn_GetMeltingTempC(GGAlign galn, int i, int j)
{
	return galn_GetMeltingTempK(galn, i,j) - 273.15f;
}


///////////////////////////////////////////////////////////////////////////////
//  FUNCTION:   float galn_GetMeltingTempK(int i, int j)
//
//  PURPOSE:    Calculate TM for given cell. Return in Kelvin.
//
//  PARAMETERS:
//      i, j    - Cell coordinates
//
//  RETURN VALUE:
//      float  - Melting Temperature TM
//
//  REVISION HISTORY
//  $              00sep06 : created LK
//  #$

float galn_GetMeltingTempK(GGAlign galn, int i, int j)
{
	float tm0, tm1, tm2;
	tm0 = nparam_CalcTM(nndS(i,j,0),nndH(i,j,0));
	tm1 = nparam_CalcTM(nndS(i,j,1),nndH(i,j,1));
	tm2 = nparam_CalcTM(nndS(i,j,2),nndH(i,j,2));
	float maxg;
	if ((tm0>=tm1)&&(tm0>=tm2))
		maxg=tm0;
	else if ((tm1>=tm0)&&(tm1>=tm2))
		maxg=tm1;
	else
		maxg=tm2;
	if (maxg<0)
		maxg=0;
	return maxg;
}


